
#import basic

from sempy.space.space import *
from sempy.basic import Basic
from sempy.space.function_vec import *
from sempy.space.plot import *
import operators
import linsolvers
import precond
import fluid
import ode
import solvers
import tests
